package ipn;

import java.io.Serializable;

/** The state of an IPN.
 * 
 * A state is comparable to other states (equals,compareTo), it can be hashed and serialized.
 * A state is non mutable ! This allows to share state references in unique tables.
 * To modify a state, use copy() first. 
 * Then update the copy using the concrete state type accessors. 
 * Finally call unique to reobtain a non mutable version of the state.
 * 
 *  Avoid calling unique on intermediate throw-away states, it triggers storage of the state in a unique table. 
 * 
 * @author yann
 *
 */
public interface IPNState extends Comparable<IPNState>, Serializable {

	/** Copy this state and obtained a mutable version of it.
	 * 
	 * @return a mutable version that is a copy of this state.
	 */
	public IPNState copy ();

	/** Produce a non-mutable version of the state.
	 * 
	 * @param type : the type of net this state belongs to. 
	 * Current implementation uses this to store states in one hashset per type, which is debatable.
	 * Parameter may be removed in future. 
	 * @return a non mutable version of this state, in most implementations should be unique, 
	 * but "return this" is acceptable trivial implementation behavior.
	 */
	public IPNState unique(IPNType type);

	/** Hash a state. */
	public int hashCode();

	/** Print a state, IPN style. */
	public String toString(IPNType net);
	
	// Implementing classes should also overload equals 
}
