package ipn;


import java.util.Collection;

import libpetri.nets.IPTNet;

/** An IPN type declaration, minimal contract.
 * 
 * A type has a name, knows a set of labeled states, and contains transitions.
 * It is able to compute successor states of a source state. 
 * It can also be exported to IPN format, and flattened for other exports to PT nets. 
 * 
 * @author yann
 *
 */
public interface IPNType {

	/** The name for this type. */
	public String getName();

	/** The labeled states of this type */
	public IPNLabeledState getState (String name);
	
	/** The transitions of this type */
	public IPNTransition getTransition (String name);
		
	/** The successors function.
	 * Return the set of successors, from the current state "source", by a multiset of transitions of this type. 
	 * @param source the source state, expected to be of type compatible with "this".
	 * @param tset The multiset of transitions to simultaneously fire.
	 * @return The set of successor states reachable in one step of this transition multiset.
	 */	
	public Collection<IPNState> getSuccessors (IPNState source, Collection<String> tset);
	
	/** The private successors function.
	 * 
	 * Returns the set of successors, from the current state "source", by firing once each private transition of this type.
	 * 
	 *  @param source the source state, expected to be of type compatible with "this".
	 *  @return The union of "getSuccessors(source, {t})" forall t in private transitions of this type.
	 *  */
	public Collection<IPNState> getPrivateSuccessors (IPNState source);
	
	/** Printing, to IPN style format. */
	public String toString();
	
	/** HashCode function. */
	public int hashCode();

	/** Flatten this net type into the receiver net.
	 * 
	 * @param receiver net to build into.
	 * @param parentname path to this instance of the type.
	 * @param state initial state of this instance of the type.
	 */
	public void flatten(IPTNet receiver, String parentname, IPNState state);
	
	/** Return the effective possible bindings for a transition multiset, in a given context.
	 * Package visibility as users should not directly call this.
	 * */
	Collection<SynchroSet> getSynchroSet (Collection<String> tset, String prefix) ;
}


