package ipn.composite;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;

import libpetri.nets.IPTNet;

import ipn.IPNInstance;
import ipn.IPNLabeledState;
import ipn.IPNModel;
import ipn.IPNState;
import ipn.IPNTransition;
import ipn.IPNType;
import ipn.ResolvedTrans;
import ipn.SynchroSet;
import ipn.Visibility;

public class CompositeNet implements IPNType, Iterable<IPNInstance> {

	private String name;
	private Vector<IPNInstance> inst = new Vector<IPNInstance>();
	private Vector<Synchronization> syncs = new Vector<Synchronization>();
	private Map<String,LabeledCompositeState> namedStates = new TreeMap<String, LabeledCompositeState>();

	/** Constructor */
	public CompositeNet(String name) {
		this.name = name ;
	}
	
	/******************** public construction API ***************************************/
	
	public boolean createInstance (String instName, String instType, IPNModel m) {
		if (getInstance(instName) != null) {
			System.err.println("Duplicate instance name "+instName+ " in composite net "+name+ " when treating createInstance("+instName+","+instType+", model)");		
			return false;
		}
		IPNType type = m.findType(instType);
		if (type == null) {
			System.err.println("Unknown type "+instType+" when treating createInstance("+instName+","+instType+", model)");
			return false;
		}
		IPNInstance i = new IPNInstance(instName,type,inst.size());
		inst.add(i);
		return true;
	}

	public boolean setSubnetState(String stateName, String instName, String substate) {
		IPNInstance i = getInstance(instName);
		if (i == null) {
			System.err.println("Unknown subnet "+instName+ " when treating setSubnetState in net type" + name + ".");
			return false;
		}
		IPNLabeledState sub = i.getType().getState(substate);
		LabeledCompositeState m = getState(stateName);		
		m.setSubstate(i, sub);
		return true;
	}
	
	public boolean createSynchronization (String syncName,Visibility vis, SyncType st) {
		if (getTransition(syncName) != null) {
			System.err.println("Synchronization "+syncName+ " already exists, when handling createSynchronization("+syncName+","+vis+","+st+")");
			return false;
		}
		Synchronization s = new Synchronization(syncName,vis,st);
		syncs.add(s);
		return true;
	}

	public boolean addSyncPart (String syncName, String instName , String slaveName) {
		Synchronization s = getTransition(syncName);
		if (s==null) {
			System.err.println("Unknown synchronization "+syncName + " when handling addSyncPart("+syncName +","+instName+","+slaveName+")");
			return false;
		}
		IPNInstance i = getInstance(instName);
		if (i==null) {
			System.err.println("Unknown instance "+instName + " when handling addSyncPart("+syncName +","+instName+","+slaveName+")");
			return false;
		}
		IPNTransition t = i.getType().getTransition(slaveName);
		if (t==null) {
			System.err.println("Unknown event "+slaveName + "in the type " + i.getType().getName() + " when handling addSyncPart("+syncName +","+instName+","+slaveName+")");
			return false;
		} else if (t.getVisibility() != Visibility.PUBLIC) {
			System.err.println("Event "+slaveName + " in the type " + i.getType().getName() + " is not PUBLIC when handling addSyncPart("+syncName +","+instName+","+slaveName+")");
			return false;
		}
		s.addPart(i,slaveName);
		return true;
	}

	/************************** Access primitives ****************************/
	
	private IPNInstance getInstance(String instName) {
		for (IPNInstance in : inst) {
			if (in.getName().equals(instName))
				return in;
		}
		return null;
	}
	
	public String getName() {
		return name;
	}
	
	public LabeledCompositeState getState(String stateName) {
		LabeledCompositeState s = namedStates.get(stateName);		
		if (s==null) {
			s= new LabeledCompositeState(this,stateName);
			namedStates.put(stateName, s);
		}
		return s;
	}

	public Synchronization getTransition(String name) {
		for (Synchronization sync : syncs){
			if (sync.getName().equals(name))
				return sync;
		}
		return null;
	}

	public Iterator<IPNInstance> iterator() {
		return inst.iterator();
	}
	
	/******************** Dynamics ********************************/
	
	private Collection<BasicSynchronization> determinize (Collection<String> tset) {
		Collection<BasicSynchronization> realSyncs = new Vector<BasicSynchronization> ();
		
		// Handle AND synchros to build a single term
		BasicSynchronization andPart = new BasicSynchronization();
		for (String tname : tset) {
			Synchronization s = getTransition(tname);
			assert s != null;
			if (s.getType() == SyncType.AND) {
				andPart.addParts(s);				
			}
		}
		realSyncs.add(andPart);
		
		// Handle OR synchros, adds one realsync per part of an or synchro.
		// /!\ multiplicative complexity in multi-OR case. 
		for (String tname : tset) {
			Synchronization s = getTransition(tname);
			if (s.getType() == SyncType.OR) {
				Collection<BasicSynchronization> newrealSyncs = new Vector<BasicSynchronization> ();
				for (Map.Entry<IPNInstance, String> part : s) {
					for (BasicSynchronization bs : realSyncs) {
						BasicSynchronization updated = new BasicSynchronization();
						updated.addParts(bs);
						updated.addPart(part.getKey(),part.getValue());
						newrealSyncs.add(updated);				
					}
				}
				realSyncs = newrealSyncs;
			}
		}
		return realSyncs;
	}
	
	public Collection<SynchroSet> getSynchroSet (Collection<String> tset, String prefix) {
		Collection<SynchroSet> ret = new Vector<SynchroSet> ();
				
		for (BasicSynchronization s : determinize(tset)) {
			assert s != null;
			// BasicSyncsAND type synchros do not produce additional entries in ret, but add their behavior to the whole.
			// get the synchro set for the current multiset of subnet transitions
			for (SynchroSet event : s.getSynchroSet(prefix)) {					
				// for each current result										
				ret.add(event);
			}
		}
	
		return ret;
	}
	
	public Collection<IPNState> getSuccessors(IPNState source,	Collection<String> tset) {
		Collection<BasicSynchronization> realSyncs = determinize(tset);
		
		// realSyncs now contains only deterministic (w.r.t this level) transitions
		Collection<IPNState> succ = new Vector<IPNState>();
		CompositeState csource = (CompositeState) source.unique(this);
		for (BasicSynchronization bs : realSyncs) {
			succ.addAll(bs.getSuccessors(csource));
		}
		return succ;
	}
	
	public Collection<IPNState> getPrivateSuccessors(IPNState source) {
		Vector<IPNState> res = new Vector<IPNState>();
		for (Synchronization s : syncs) {
			if (s.getVisibility() == Visibility.PRIVATE) {
				Vector<String> tset = new Vector<String> ();
				tset.add(s.getName());
				for (IPNState succ : getSuccessors(source.copy(), tset))
					res.add(succ.unique(this));
			}
		}
		CompositeState src = (CompositeState) source;
		for (IPNInstance i: inst) {
			Collection<IPNState> localsucc = i.getType().getPrivateSuccessors(src.getSubstate(i));
			for (IPNState loc : localsucc ) {
				CompositeState s2 = src.copy();
				s2.setSubstate(i, loc);
				res.add(s2.unique(this));
			}
		}
		return res;
	}
		
	/****************** Export functions *****************************/
	
	public void flatten(IPTNet receiver, String parentname, IPNState state) {
		for (IPNInstance inst:this) {
			inst.getType().flatten(receiver,parentname + "." + inst.getName(), ((CompositeState)state).getSubstate(inst) );
		}
		for (Synchronization s: syncs) {
			if (s.getVisibility() == Visibility.PRIVATE) {
				Vector<String> tset = new Vector<String> ();
				tset.add(s.getName());
				Collection<SynchroSet> setsyncs = getSynchroSet(tset,parentname); 
				int n=0;
				for (SynchroSet ssync : setsyncs) {
					String restname = parentname+"."+s.getName()+ n++;
					receiver.createTransition(restname, libpetri.structure.Visibility.PRIVATE);
					for (ResolvedTrans t : ssync) {					
						t.getTrans().flatten(receiver, restname  ,t.getPath());
					}
					
				}
			}
		}
	}
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("\nCompositeNet "+name+ " {\n");
		sb.append("\n\t// subnet declarations\n");
		for (IPNInstance sub:inst) {
			sb.append("\tsubnet "+sub.getType().getName() +" "+ sub.getName() + " ;\n");
		}
		sb.append("\n\t// Named states declarations\n");
		for (IPNLabeledState st: namedStates.values()) {
			sb.append(st.toString(this));
		}
		sb.append("\n\t// Synchronization declarations\n");
		for (Synchronization t: syncs) {
			sb.append(t);
		}
		sb.append("}\n");
		return sb.toString();
	}


	/***************** trivia **************************/
	public int hashCode() {
		return name.hashCode() ^ inst.hashCode() ^ syncs.hashCode();
	}

	
}



