package ipn.composite;

import ipn.IPNInstance;
import ipn.IPNState;
import ipn.IPNType;


import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Vector;


/** A not-so basic implementation of a net state.
 * 
 * @author yann
 * 
 * Contains a beautiful hidden mechanism to ensure state uniqueness at any depth.
 * Protocol : get/copy/set* /unique ;
 *  * NetState.copy ->Helper and Helper.unique ->NetState
 * set on Uniquehelper handles dirty flags and late copy of data.
 * unique relies on a static table ! BE WARNED THIS IS NOT thread-safe
 * if you want thread safe basic implem of states, look at revision 1610. 
 * or just set unique flag to false
 */
public class CompositeState implements IPNState,Cloneable,Serializable {

	//	if unique then use hierarchical unicity table + late copy mechanism
	//  else use a basic copy mechanism
	private static final boolean UNIQUE = true;
	/// to store the states of subnets
	protected Vector<IPNState> comp;
	
	/// initialize sizes w.r.t. n
	public CompositeState(CompositeNet net) {		
		comp = new Vector<IPNState>();
		for (IPNInstance i : net) {
			comp.add(i.getType().getState("default"));
		}
		comp.trimToSize();
	}
	/// for use by UniqueHelper
	protected CompositeState(Vector<IPNState> comp) {
		this.comp = comp;
		comp.trimToSize();
	}
	/// public copy semantics constructor
	public CompositeState(CompositeState state) {
		comp = new Vector<IPNState>();
		for (IPNState s : state.comp) {
			comp.add(s.copy());
		}
		comp.trimToSize();
	}
	/** modify place state. 
	 * Size checks are made.
	 * 
	 */
	public void setSubstate(IPNInstance inst, IPNState sub) {
		int i = inst.getIndex();
		assert (this instanceof UniqueHelper);
		if (i >= comp.size()) 
			comp.setSize(i+1);		
		comp.set(i,sub);		
	}
	
	public IPNState getSubstate(IPNInstance key) {
		int i = key.getIndex();
		IPNState s =  getSubstate (i);
		if (s != null)
			return s;
		return key.getType().getState("default");
	}
	
	protected IPNState getSubstate (int index) {
		if (index < comp.size()) {
			IPNState s = comp.get(index); 
			return s;
		}
		return null;		
	}
	

	public CompositeState copy() {
		if (UNIQUE)
			return new UniqueHelper(this);
		else 
			return new CompositeState(this);
	}

	public String toString() {
		return  comp.toString();
	}

	public String toString (IPNType n) {
		CompositeNet net = (CompositeNet) n;
		StringBuffer sb = new StringBuffer();		
		for (IPNInstance inst : net) {
			String s = getSubstateString(inst);
			if (! s.equals("default"))
				sb.append("\t  "+inst.getName()+ " = "+ s +" ;\n");
		}
		return sb.toString();
	}

	
	protected String getSubstateString(IPNInstance inst) {
		return getSubstate(inst).toString(inst.getType());
	}
	@Override
	public int hashCode() {
		int ret = 1153;
		for (IPNState s : comp){
			ret ^= 31*ret + (s == null ? 0 : s.hashCode()) + 7119;
		}
		return ret;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (obj instanceof CompositeState){
			final CompositeState other =  (CompositeState) obj;
			if (comp == null) {
				if (other.comp != null)
					return false;
			} else if (!comp.equals(other.comp))
				return false;
		} else {
			return false;
		}
		return true;
	}

	public IPNState unique(IPNType net) {
		if (UNIQUE)
			return new UniqueHelper(this).unique(net);
		else 
			return this;
	}

	public static String getUnicityStats () {
		return UniqueHelper.toStringUniqueTable();
	}
	public int compareTo(IPNState s2) {		
		int ret=0;
		if (s2 instanceof CompositeState) {
			CompositeState s = (CompositeState) s2;
			// check component vector size
			ret = comp.size() - s.comp.size();
			if (ret != 0) {
				return ret/ Math.abs(ret);
			}
					
			// lexico on components
			for (int i=0;i < comp.size();i++) {
				ret = comp.get(i).compareTo(s.comp.get(i));
				if (ret !=0)
					return ret;				
			}
		} else {
			return -1;
		}
		return ret;
	}
}

class UniqueHelper extends CompositeState {
	/// stores : net -> states set
	private static HashMap<IPNType, HashMap<CompositeState,CompositeState>> states_ = new HashMap<IPNType, HashMap<CompositeState,CompositeState>> ();

	//	 true if comp vector has been modified already
	private boolean isDirtyComp;
	// true cell per cell if the component has been touched. Implies isDirtyComp.
	private boolean[] isDirtyTab;

	public UniqueHelper(CompositeState state) {
		// NO COPY using protected constructor
		super(state.comp);
		// everyone is clean
		isDirtyComp = false;
		
	}


	@Override
	public CompositeState copy() {
		UniqueHelper build = new UniqueHelper(this);
		if (isDirtyComp) {
			build.comp = new Vector<IPNState> (comp);
			build.isDirtyComp = true;
			build.isDirtyTab = isDirtyTab.clone();
			for (int i = 0; i < comp.size() ; ++i) {
				if (isDirtyTab[i]) {
					build.comp.set(i, comp.get(i).copy());
				}
			}
		}
		return build;
	}
	
	public CompositeState unique(IPNType type) {
		CompositeNet net = (CompositeNet) type;
		// get net states set
		HashMap<CompositeState, CompositeState> st = states_.get(net);
		// first hit for this net type
		if (st == null) {
			st = new HashMap<CompositeState, CompositeState>();
			states_.put(net, st);
		}
		// check if components should uniquified
		if (isDirtyComp) {
			int i=0;
			for (IPNInstance inst : net) {
				if (isDirtyTab[i]) {
					// dirty components get cleansed by a unique.
					if (! comp.get(i).equals(inst.getType().getState("default")))
						comp.set(i, comp.get(i).unique(inst.getType()));
				}
			}
		}
		// should really uniquify the place vect to be fully homogeneous, but it's int[]
		// not really an issue unless user mixes standard places/trans and subcomponents frequently
		// using (nets = composites) + (subnets = basic) => no overhead by not uniqueifying marks

		// dummy state, trimming all the extra dirty bits from this.		
		CompositeState s = new CompositeState (comp);
		// lookup
		synchronized (st) {
			CompositeState concrete = st.get(s);
			if (concrete == null) {
				// new state
				st.put(s,s);
				return s;
			} else {
				// known state
				return concrete;
			}
		}
	}

	@Override
	// set the dirty flags and copy the data just before modifying
	public void setSubstate(IPNInstance inst, IPNState sub) {
		int i = inst.getIndex();
		if (! isDirtyComp || i >= comp.size()) {
			isDirtyComp=true;
			if (i >= comp.size()) {				
				comp = new Vector<IPNState>(comp);
				comp.setSize(i+1);
				
				boolean[] oldDirt = isDirtyTab;
				// filled with default java boolean = false
				isDirtyTab = new boolean [i+1];
				if (oldDirt != null) 
					System.arraycopy(oldDirt, 0, isDirtyTab, 0, oldDirt.length);
			} else {
				comp = new Vector<IPNState>(comp);
				// filled with default java boolean = false
				isDirtyTab = new boolean [comp.size()+1];
			}			
		}
		if (! isDirtyTab[i]) {
			isDirtyTab[i] = true;
	//		comp.set(i, getSubstate(inst).copy());
		}
		super.setSubstate(inst, sub);
	}

	// for debug, pretty print of unique tables
	public static String toStringUniqueTable() {
		StringBuffer sb = new StringBuffer();
		sb.append("Unicity table : " + states_.size() + " net types seen.\n");
		for (IPNType netType : states_.keySet()) {
			HashMap<CompositeState, CompositeState> states = states_.get(netType);
			sb.append(">>>Net :"+netType.getName() + " states :"+ states.size()+"\n");
			int j = 0;
			Vector<CompositeState> keys = new Vector<CompositeState>(states.keySet());
			Collections.sort(keys);
			for (CompositeState s : keys) {
				// would like s.toString(net), but which net ??? we only have the class and no instance.
				sb.append("\t"+j+++":"+s.toString()+"\n");
			}
		}		
		return sb.toString();		
	}


}

