package ipn.elementary;

import java.util.Collection;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;

import libpetri.nets.IPTNet;

import ipn.IPNLabeledState;
import ipn.IPNState;
import ipn.IPNType;
import ipn.ResolvedTrans;
import ipn.SynchroSet;
import ipn.Visibility;

public class UnitaryType implements IPNType {

	Vector<Place> places = new Vector<Place>();
	Vector<UnitaryTransition> transitions = new Vector<UnitaryTransition>();
	private String name;
	private Map<String,IPNLabeledState> namedStates = new TreeMap<String, IPNLabeledState>();


	public UnitaryType(String name) {
		this.name = name ;
	}

	public String getName() {
		return name;
	}

	public LabeledMarking getState(String stateName) {
		LabeledMarking m = (LabeledMarking) namedStates.get(stateName);
		if (m==null) {
			m = new LabeledMarking(stateName,this);
			namedStates.put(stateName, m);
		}		
		return m;
	}


	public Place getPlace(String name) {
		for (Place p : places) {
			if (p.getName().equals(name))
				return p;
		}
		return null;
	}

	public UnitaryTransition getTransition(String name) {
		for (UnitaryTransition t : transitions) {
			if (t.getName().equals(name))
				return t;
		}
		return null;
	}


	public boolean createPlace(String name) {
		if (getPlace(name) != null) {
			System.err.println("Place "+name+ " already exists when treating createPlace in net "+this.name);
			return false;
		} else {
			places.add(new Place(name,places.size()));
			return true;
		}
	}

	public boolean createTransition(String name, Visibility vis) {
		if (getTransition(name) != null) {
			System.err.println("Transition "+name+ " already exists when treating createTransition in net "+this.name);
			return false;
		} else {
			transitions.add(new UnitaryTransition(name,vis));
			return true;
		}
	}

	public boolean createArc(String pname, String tname, ArcType at, int value) {
		Place p = getPlace(pname);
		if (p == null) {
			System.err.println("Unknown place "+pname+ " when treating createArc("+pname+","
					+tname+","+at+")");
			return false;
		}
		UnitaryTransition t = getTransition(tname);
		if (t == null) {
			System.err.println("Unknown transition "+tname+ " when treating createArc("+pname+","
					+tname+","+at+")");
			return false;
		}
		t.addArc(p,at,value);
		return true;
	}

	public boolean setPlaceState(String stateName, String placeName, int mark) {
		Place p = getPlace(placeName);
		if (p == null) {
			System.err.println("Unknown place "+placeName+ " when treating setPlaceState in net type" + name + ".");
			return false;
		}
		LabeledMarking m = getState(stateName);
		m.setPlaceState(p, mark);
		return true;
	}
	
	private boolean isEnabled(IPNState state, Collection<String> tset) {
		Marking m = (Marking) state;
		for (String it : tset) {			
			UnitaryTransition t = getTransition(it);
			if (t.isInhibited(m))
				return false;
		}
		m = m.copy();
		for (String it : tset) {			
			UnitaryTransition t = getTransition(it);
			// updates the state to have correct results in multiset case
			if (! t.isInputEnabled(m))
				return false;
		}
		return true;
	}


	private void fire(Marking source, String tname) {
		UnitaryTransition t = getTransition(tname);
		assert t != null;
		t.fire(source);
	}
	

	public Collection<SynchroSet> getSynchroSet(Collection<String> tset, String prefix) {
		Collection<SynchroSet> ret = new Vector<SynchroSet>();
		SynchroSet toAdd = new SynchroSet();
		for (String tname : tset) {
			toAdd.add(new ResolvedTrans(prefix,getTransition(tname)));
		}
		ret.add(toAdd);
		return ret;
	}
	
	public void flatten(IPTNet receiver, String parentname, IPNState state) {
		Marking mark = (Marking) state;
		for (Place p : places) {
			String n = parentname +"."+ p.getName();
			receiver.createPlace(n);
			receiver.setPlaceState(n, mark.getPlaceState(p));
		}
		for (UnitaryTransition t : transitions) {
			if (t.getVisibility() == Visibility.PRIVATE) {
				String resname = parentname+"."+t.getName();
				receiver.createTransition(resname, libpetri.structure.Visibility.PRIVATE);
				t.flatten(receiver,resname,parentname);
			}
		}		
	}

	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("\nUnitaryNet "+name+ " {\n");
		sb.append("\n\t// Place declarations\n");
		for (Place p : places) {
			sb.append(p);
		}
		sb.append("\n\t// Named states declarations\n");
		for (IPNLabeledState st: namedStates.values()) {
			sb.append(st.toString(this));
		}
		sb.append("\n\t// Transition declarations\n");
		for (UnitaryTransition t: transitions) {
			sb.append(t);
		}
		sb.append("}\n");
		return sb.toString();
	}

	public Collection<IPNState> getSuccessors(IPNState source,	Collection<String> tset) {
		if (isEnabled(source, tset)) {
			Marking m = (Marking) source.copy();
			for (String tname : tset)
				fire(m,tname);
			
			Vector<IPNState> ret = new Vector<IPNState>();
			ret.add(m);
			return ret;
		} else {
			// empty set
			return new Vector<IPNState>();
		}
	}
	
	public Collection<IPNState> getPrivateSuccessors(IPNState source) {
		Vector<IPNState> res = new Vector<IPNState>();
		for (UnitaryTransition t : transitions) {
			if (t.getVisibility() == Visibility.PRIVATE) {
				Vector<String> tset = new Vector<String> ();
				tset.add(t.getName());
				res.addAll(getSuccessors(source, tset));
			}
		}
		for (int i = 0; i < res.size(); i++) {
			res.set(i, res.get(i).unique(this));
		}
		return res;
	}


}
