package libpetri.structure.state;

import interfaces.IState;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Vector;

import libpetri.structure.Net;
import libpetri.structure.Place;
import libpetri.tensor.TensorSet;

/** A basic implem of a net state.
 * 
 * @author yann
 * 
 * Contains a beautiful hidden mechanism to ensure state uniqueness at any depth.
 * Protocol : get/copy/set* /unique ;
 *  * NetState.copy ->Helper and Helper.unique ->NetState
 * set on Uniquehelper handles dirty flags and late copy of data.
 * unique relies on a static table ! BE WARNED THIS IS NOT thread-safe
 * if you want thread safe basic implem of states, look at revision 1610. 
 * or just set unique flag to false
 */
public class NetState implements State,Cloneable,IState,Serializable {

	//	if unique then use hierarchical unicity table + late copy mechanism
	//  else use a basic copy mechanism
	private static final boolean UNIQUE = true;
	/// to store the states of subnets
	protected Vector<State> comp;
	/// to store place states
	protected TensorSet marks;
	
	/// initialize sizes w.r.t. n
	public NetState(Net n) {
		marks = new TensorSet();
		for (Place p: n.getPlaces()) {
			marks.addPlace(p);
		}
		comp = new Vector<State>();
	}
	/// for use by UniqueHelper
	protected NetState(Vector<State> comp,TensorSet marks) {
		this.comp = comp;
		this.marks = marks;
	}
	/// public copy semantics constructor
	public NetState(NetState state) {
		marks = state.marks.copy();
		comp = new Vector<State>();
		for (State s : state.comp) {
			comp.add(s.copy());
		}
	}
	/** modify place state. 
	 * Size checks are made.
	 * 
	 */
	public void addToPlaceState(Place p,PlaceState value) {		
//		marks.ensureRankSize(0,index+1);
//		Tensor t2 = value.t.insert(0, marks.getRankSize(0), index);
//		System.out.println("nb places : " +marks.getRankSize(0));
//		
//		System.out.println("add : " +t2);
//		System.out.println("and : " +marks);
		if (marks.size()==p.getIndex())
			marks.addPlace(p);
		marks.add(p.getIndex(),value.t);
//		System.out.println("obtained : " +marks);
		
	}
	/// modify place state. Size checks are made.
	public void removeFromPlaceState(Place p,PlaceState value) {		
		//marks.ensureRankSize(0,index);
		// Tensor t2 = value.t.insert(0, marks.getRankSize(0), index);
		marks.remove(p.getIndex(),value.t);
	}
	
	public State copy() {
		if (UNIQUE)
			return new UniqueHelper(this);
		else 
			return new NetState(this);
	}

	public PlaceState getPlaceState(Place p) {
		return new PlaceState(marks.get(p.getIndex()));
	}

	public String toString() {
		return  marks.toString() + comp;
	}
	
	public String toString (Net n) {
		StringBuffer sb = new StringBuffer("\n");
		for (Place p:n.getPlaces()) {
			PlaceState m = getPlaceState(p);
			if (! m.isEmpty())
				sb.append("\t"+p.getName()+" = "+m.toString(p.getDom()) +";");
		}
		int i =0;
		for (Net c : n.getComponents()) {
			sb.append(comp.get(i++).toString(c));
		}
		return sb.toString();
	}

	public State getComponentState(int i) {
		return comp.get(i);
	}
	
	public void setComponentState(int i,State s) {
		if (i >= comp.size()) 
			comp.add(i,s);
		else
			comp.set(i,s);
	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + ((comp == null) ? 0 : comp.hashCode());
		result = PRIME * result + marks.hashCode();
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final NetState other = (NetState) obj;
		if (comp == null) {
			if (other.comp != null)
				return false;
		} else if (!comp.equals(other.comp))
			return false;
		if (!marks.equals(other.marks))
			return false;
		return true;
	}

	public State unique(Net net) {
		if (UNIQUE)
			return new UniqueHelper(this).unique(net);
		else 
			return this;
	}
	
	public static String getUnicityStats () {
		return UniqueHelper.toStringUniqueTable();
	}
	public int compareTo(State s2) {		
		int ret=0;
		if (s2 instanceof NetState) {
			NetState s = (NetState) s2;
			// check component vector size
			ret = comp.size() - s.comp.size();
			if (ret != 0) {
				return ret/ Math.abs(ret);
			}
			// check place vector size
			ret = marks.size() - s.marks.size();
			if (ret != 0) {
				return ret/ Math.abs(ret);
			}		
			// lexico on marks
			ret = marks.compareTo(s.marks); 	
			if (ret !=0)
				return ret;
			// lexico on components
			for (int i=0;i < comp.size();i++) {
				ret = comp.get(i).compareTo(s.comp.get(i));
				if (ret !=0)
					return ret;				
			}
		}
		return ret;
	}
}

class UniqueHelper extends NetState {
	/// stores : net -> states set
	private static HashMap<Class, HashMap<State,State>> states_ = new HashMap<Class, HashMap<State,State>>();
	
	// true if place marks has been touched
	private boolean isDirtyMarks;
	//	 true if comp vector has been modified already
	private boolean isDirtyComp;
	// true cell per cell if the component has been touched. Implies isDirtyComp.
	private boolean[] isDirtyTab;

	public UniqueHelper(NetState state) {
		// NO COPY using protected constructor
		super(state.comp, state.marks);
		// everyone is clean
		isDirtyComp = false;
		isDirtyMarks = false;
		// filled with default java boolean = false
		isDirtyTab = new boolean [comp.size()];
	}

	public State unique(Net net) {
		// get net type
		Class netType = net.getClass();
		// get net states set
		HashMap<State, State> st = states_.get(netType);
		// first hit for this net type
		if (st == null) {
			st = new HashMap<State, State>();
			states_.put(netType, st);
		}
		// check if components should uniquified
		if (isDirtyComp) {
			int i=0;
			for (Net n : net.getComponents()) {
				if (isDirtyTab[i]) {
					// dirty components get cleansed by a unique.
					comp.set(i, comp.get(i).unique(n));
				}
			}
		}
		// should really uniquify the place vect to be fully homogeneous, but it's int[]
		// not really an issue unless user mixes standard places/trans and subcomponents frequently
		// using (nets = composites) + (subnets = basic) => no overhead by not uniqueifying marks
		
		// canonicity 
		// test for 12 clients
//		if (comp.size() == 13) {
//			Collections.sort(comp.subList(1, 13));
//		}
		
		// dummy state, trimming all the extra dirty bits from this.		
		State s = new NetState (comp,marks);
		// lookup
		synchronized (st) {
		State concrete = st.get(s);
		if (concrete == null) {
			// new state
			st.put(s,s);
			return s;
		} else {
			// known state
			return concrete;
		}
		}
	}
	
	@Override
	// set the dirty flags and copy the data just before modifying
	public void setComponentState(int i, State s) {
		if (! isDirtyComp) {
			isDirtyComp=true;
			comp = new Vector<State>(comp);
		}
		if (! isDirtyTab[i]) {
			isDirtyTab[i] = true;
			comp.set(i, comp.get(i).copy());
		}
		super.setComponentState(i, s);
	}
	
	@Override
	// set the dirty flags and copy the data just before modifying
	public void addToPlaceState(Place p, PlaceState value) {
		if (!isDirtyMarks) {
			isDirtyMarks = true;
			marks = marks.copy(); 
		}
		super.addToPlaceState(p, value);
	}
	// set the dirty flags and copy the data just before modifying
	public void removeFromPlaceState(Place p, PlaceState value) {
		if (!isDirtyMarks) {
			isDirtyMarks = true;
			marks = marks.copy(); 
		}
		super.removeFromPlaceState(p, value);
	}

	
	// for debug, pretty print of unique tables
	public static String toStringUniqueTable() {
		StringBuffer sb = new StringBuffer();
		sb.append("Unicity table : " + states_.size() + " net types seen.\n");
		for (Class netType : states_.keySet()) {
			HashMap<State, State> states = states_.get(netType);
			sb.append(">>>Net :"+netType.getSimpleName() + " states :"+ states.size()+"\n");
			int j = 0;
			for (State s : states.keySet()) {
				// would like s.toString(net), but which net ??? we only have the class and no instance.
				sb.append("\t"+j+++":"+s.toString()+"\n");
			}
		}		
		return sb.toString();		
	}
	
}