package transform;


import ipn.IPNModel;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import libpetri.ami.AMInet;
import libpetri.dot.DOTnet;
import libpetri.structure.Net;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.uml2.uml.Activity;
import org.eclipse.uml2.uml.Behavior;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Interaction;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.resource.UMLResource;

import transform.activity.ActivityTransformer;
import transform.activity.ActivitytoIPNTransformer;
import transform.activity.ActivitytoNetTransformer;
import transform.sd.InteractionTransformer;
import transform.sd.SDToStringTransformer;
import transform.stm.STMtoNetTransformer;
import transform.stm.StateMachineTransformer;



public class SimpleTransformer {
	private Trace <Element,String> trace= new Trace<Element, String>();
	private Map<NamedElement, Net> products = new HashMap<NamedElement, Net>();

	private InteractionTransformer interactionTransf = new SDToStringTransformer ();
	private StateMachineTransformer stmTransf = new STMtoNetTransformer (trace, products);
	private ActivityTransformer actTransf = new ActivitytoNetTransformer (trace, products);
	
	
	public Map<NamedElement, Net> transform (Configuration conf) {
		// load model from URI
		String modelff = conf.getModelPath();
		URI fileURI = URI.createFileURI(modelff);

		// Set up a resource
		ResourceSet resourceSet = new ResourceSetImpl();
		// UML specific ecore setup
		resourceSet.getPackageRegistry().put(UMLPackage.eNS_URI, UMLPackage.eINSTANCE);
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(UMLResource.FILE_EXTENSION, UMLResource.Factory.INSTANCE);

		//Create a Resource in order to manipulate the process model instance
		// this call actually loads the model into memory
		Resource resource = resourceSet.getResource(fileURI, true);

		return handleModel(resource,conf);
	}


	private Map<NamedElement, Net> handleModel(Resource resource, Configuration conf) {
		// Basic traversal
		for (Object o : resource.getContents()) {
			// treat all top level elements of the resource.
			if (o instanceof Element) {
				System.out.println("Found a top level model element: \n" + o);
				Element me = (Element) o;
				handleModelElement(me,conf);
			} else 
				System.out.print('.');
			//			System.out.println("Top level resource is not a UML element : \n" + o);
		}
		
		System.out.println(interactionTransf);
		
		return products;
	}



	public Map<NamedElement, Net> handleModel(Model model, Configuration conf) {
		for ( Element e : model.allOwnedElements()) {					
			handleModelElement(e,conf);
		}
		return products;
	}


	public Map<NamedElement, Net> handleModelElement(Element e, Configuration conf) {
		//					System.out.println(e);
		if (e instanceof StateMachine) {
			StateMachine machine = (StateMachine) e;
			stmTransf.handleSTM(machine,conf);
		} else if (e instanceof Activity) {
			Activity act = (Activity) e;
			actTransf.handleActivity (act,conf);
		} else if (e instanceof org.eclipse.uml2.uml.Package) {
			// recurse into packages 
			org.eclipse.uml2.uml.Package pkg = (org.eclipse.uml2.uml.Package) e;
			for ( Element elt : pkg.allOwnedElements()) {
				handleModelElement(elt,conf);
			}
		} else if (e instanceof Model) {
			// go one level deeper into the model
			Model model = (Model) e;
			for ( Element ee : model.allOwnedElements()) {					
				handleModelElement(ee,conf);
			}
		} else if (e instanceof Interaction) {
			Interaction in = (Interaction) e;
			interactionTransf.handleInteraction (in,conf);
		} else {
			System.out.print('.');
			//System.err.println("Unparsed model element :" + e);
		}
		// else if SequenceDiagram ... etc 
		return products;
	}


	@SuppressWarnings("unused")
	private static void exhibit(Net net, Configuration conf) {
		System.out.println(net);
		try {
			DOTnet a = new DOTnet(conf.getModelName()+"_"+net.getName());
			net.convert(a);
			a.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		try {
			AMInet a = new AMInet();
			net.convert(a);
			a.writeNet(conf.getTempCamiPath());
		} catch (IOException e) {
			e.printStackTrace();
		}
	}


	public Trace<Element, String> getTrace() {
		return trace;
	}

	public void setActivityTransformer(ActivityTransformer actTransf) {
		this.actTransf = actTransf;
	}
	
	public static void main(String[] args) {
		Configuration conf = new Configuration();
		conf.setModelPath("ModelFolder/models/simple/BCC-example.uml");
		SimpleTransformer st = new SimpleTransformer();
		IPNModel m = new IPNModel();
		BehaviorTransformationTrace btt = new BehaviorTransformationTrace();
		st.setActivityTransformer (new ActivitytoIPNTransformer(st.trace,m,btt));
		st.transform(conf);
		
		/// now handle composition
		
		// basic case : test each diagram in isolation.
		for (Entry<Behavior, IPNBehavior> it : btt.entrySet()) {
			IPNTransformer.createIsolated(m, it.getValue());
			m.setMain(IPNTransformer.getIsolatedName(it.getValue().getTypeName()), "main");
			exhibitModel(m);
		}
		// compose with depth 1
		for (Entry<Behavior, IPNBehavior> it : btt.entrySet()) {
			IPNTransformer.createConnected(m, it.getValue(), btt, 1);
			m.setMain(IPNTransformer.getConnectedName(it.getValue().getTypeName(),1), "main");
			exhibitModel(m);
		}
		
		System.out.println(m);
		
		
		m.setMain("BCC-example::Shipping", "main");
		exhibitModel(m);
		

	}


	private static void exhibitModel(IPNModel m) {
		DOTnet net;
		String name = m.getMain().getType().getName();
		String path = "ModelFolder/models/simple/";
		try 	{
			net = new DOTnet(name,path);
			m.flatten(net);
			net.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}


}
