package transform.sd;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.CombinedFragment;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Interaction;
import org.eclipse.uml2.uml.InteractionConstraint;
import org.eclipse.uml2.uml.InteractionFragment;
import org.eclipse.uml2.uml.InteractionOperand;
import org.eclipse.uml2.uml.InteractionUse;
import org.eclipse.uml2.uml.Lifeline;
import org.eclipse.uml2.uml.LiteralInteger;
import org.eclipse.uml2.uml.MessageOccurrenceSpecification;
import org.eclipse.uml2.uml.OpaqueExpression;
import org.eclipse.uml2.uml.ValueSpecification;

import transform.Configuration;

public class SDToStringTransformer implements InteractionTransformer {
	// a map : "instance:class" -> ( interaction -> "formula")
	// each lifeline is mapped to one formula per interaction
	private Map<String, Map<Interaction,String> > lifelines = new HashMap<String, Map<Interaction,String> >();

	
	public void handleInteraction(Interaction inter, Configuration conf) {
		// initialize all lifeline maps if needed
		for (Element e2 : inter.allOwnedElements()) {
			if (e2 instanceof Lifeline) {
				Lifeline l = (Lifeline) e2;
				String lname = getLifeName(l);
				if (! lifelines.containsKey(lname)) {
					lifelines.put(lname, new HashMap<Interaction, String> ());
				}
			}
		}
		// compute one formula for each lifeline of the interaction
		for (Lifeline l : inter.getLifelines()) {
			String lname = getLifeName(l);
			if (! lifelines.containsKey(lname)) {
				lifelines.put(lname, new HashMap<Interaction, String> ());
			}
			Map<Interaction, String> formulas = lifelines.get(lname);

			if (! formulas.containsKey(inter) ) {
				String formula = "";
				for (InteractionFragment itf : inter.getFragments()) {
					for (Lifeline l2 : itf.getCovereds()) {
						if (lname.equals(getLifeName(l2))) {
							formula += handleInteractionFragment(itf, conf, "", lname);
						}
					}
				}
				formulas.put(inter, formula);
			}
		}		
	}

	private static String getLifeName(Lifeline l) {
		return l.getRepresents().getName()+":"+l.getRepresents().getType().getName();
	}


	private String handleInteractionFragment(InteractionFragment t, Configuration conf, String piecestr, String lifeline) {
		StringBuffer piece = new StringBuffer(piecestr);
		
		if (t instanceof MessageOccurrenceSpecification) {
			MessageOccurrenceSpecification mos = (MessageOccurrenceSpecification) t;

			for (Lifeline lf : mos.getCovereds()) { 
				 // the occurrence is treated only if its owner is "lifeline"
				if (lifeline.equals(getLifeName(lf))) { 
					if (mos.getEvent().getClass().getName().contains("Send")) {
						piece.append("Send("+mos.getMessage().getName()+"(");
						// handle send arguments.
						appendMessageArguments (piece,mos.getMessage().getArguments());						
						// close the sendMessage
						piece.append("),"+mos.getMessage().getMessageSort()+") ; ");
					} else if (mos.getEvent().getClass().getName().contains("Receive")){
						piece.append("Receive("+mos.getMessage().getName()+"(");
						// handle arguments.
						appendMessageArguments (piece,mos.getMessage().getArguments());		
						// close it
						piece.append("),"+mos.getMessage().getMessageSort()+") ; ");
					} else if (mos.getEvent().getClass().getName().contains("Creation") || mos.getEvent().getClass().getName().contains("Destruction")){
						// create/destruct instance
						piece.append(mos.getMessage().getMessageSort()+"("+mos.getMessage().getName()+") ; ");
					} else {
						// ???
						piece.append(mos.getEvent().getClass().getName()+"("+mos.getMessage().getName()+","+mos.getMessage().getMessageSort()+") ; ");
					}
				}
			}
		} else if (t instanceof InteractionUse) {
			InteractionUse iu = (InteractionUse) t;
			handleInteraction(iu.getRefersTo(), conf);			
			piece.append( lifelines.get(lifeline).get(iu.getRefersTo()));
		} else if (t instanceof CombinedFragment) {
			CombinedFragment cf = (CombinedFragment) t;
			piece.append(cf.getInteractionOperator()+"%");
			for (Element cfe : cf.getOperands()) {
				InteractionOperand io = (InteractionOperand) cfe;
				InteractionConstraint ic = io.getGuard();
				piece.append("{ [");
				for (Element ice : ic.getOwnedElements()) {
					if (ice instanceof OpaqueExpression) {
						OpaqueExpression oei = (OpaqueExpression) ice;
						for (String oei_elt : oei.getBodies()) {
							piece.append(oei_elt+" ");
						}
					} else if (ice instanceof LiteralInteger) {
						LiteralInteger lii = (LiteralInteger) ice;
						piece.append(lii.getValue()+" ");
					} else {
						piece.append(ice.toString());
					}
				}
				piece.append("] ");
				for (InteractionFragment ioe : io.getFragments()) {
					String ifInit = "";
					piece.append(handleInteractionFragment(ioe, conf, ifInit, lifeline));
				}
				piece.append(" } ");
			}
			piece.append("% ; ");
		}
		return piece.toString();
	}


	private void appendMessageArguments(StringBuffer piece,
			EList<ValueSpecification> args) {
		if (! args.isEmpty()) {
			for (ValueSpecification iter : args) {
				piece.append(iter.getName());
				piece.append(",");
			}
			// remove extra comma generated by above loop.
			piece.deleteCharAt(piece.length()-1);
		}
	}


	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for ( Map.Entry<String, Map<Interaction,String>> elt : lifelines.entrySet()) {
			sb.append("\nLifeline "+elt.getKey()+"\n");
			for ( Map.Entry<Interaction,String> formula : elt.getValue().entrySet()) {
				sb.append("   inter: " + formula.getKey().getQualifiedName()+"\n" 
						 + formula.getValue()+"\n");
			}
		}
		return sb.toString();
	}
	
}
